#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>
#include <afblib/hostport.h>
#include <afblib/outbuf_printf.h>
#include "mxprequest.h"
#include "mxpresponse.h"

int main(int argc, char** argv) {
   const char* cmdname = *argv++; --argc;
   const char usage[] = "Usage: %s hostport lock command\n";
   outbuf err = {2};
   if (argc < 3) {
      outbuf_printf(&err, usage, cmdname);
      outbuf_flush(&err);
      exit(1);
   }
   const char* hptext = *argv++; --argc;
   const char* lockname = *argv++; --argc;

   /* check validity of lockname which ought to be an
      arbitrary sequence of 8-bit characters
      with the exception of `/', CR, LF, and null-bytes */
   for (const char* cp = lockname; *cp; ++cp) {
      if (*cp == '/' || *cp == '\r' || *cp == '\n') {
	 outbuf_printf(&err, "%s: invalid lockname\n", cmdname, lockname);
	 outbuf_flush(&err);
	 exit(1);
      }
   }

   /* connect to server */
   hostport hp;
   if (!parse_hostport(hptext, &hp, 21021)) {
      outbuf_printf(&err, "%s: unable to parse hostport: %s\n",
	 cmdname, hptext);
      outbuf_flush(&err);
      exit(1);
   }
   int fd;
   if ((fd = socket(hp.domain, SOCK_STREAM, hp.protocol)) < 0) {
      outbuf_printf(&err, "%s: socket: %s\n", cmdname, strerror(errno));
      outbuf_flush(&err);
      exit(1);
   }
   if (connect(fd, (struct sockaddr *) &hp.addr, hp.namelen) < 0) {
      outbuf_printf(&err, "%s: connect to %s: %s\n",
	 cmdname, hptext, strerror(errno));
      outbuf_flush(&err);
      exit(1);
   }

   inbuf mxp_in = {fd}; outbuf mxp_out = {fd};
   mxp_response response = {0};
   mxp_request request = {{0}};

   /* read initial greeting */
   if (!read_mxp_response(&mxp_in, &response) ||
	 response.status != MXP_SUCCESS) {
      outbuf_printf(&err, "%s: got no greeting from %s\n", cmdname, hptext);
      outbuf_flush(&err);
      exit(1);
   }

   /* identify ourselves */
   ssize_t namelen = sysconf(_SC_HOST_NAME_MAX) + 1;
   char* hostname = malloc(namelen);
   if (!hostname || gethostname(hostname, namelen) < 0) {
      outbuf_printf(&err, "%s: gethostname failed\n", cmdname);
      outbuf_flush(&err);
      exit(1);
   }
   stralloc myname = {0};
   stralloc_copys(&myname, hostname);
   stralloc_cats(&myname, "[");
   stralloc_catlong0(&myname, getpid(), 1);
   stralloc_cats(&myname, "]");
   stralloc_copys(&request.keyword, "id");
   stralloc_copy(&request.parameter, &myname);
   if (!write_mxp_request(&mxp_out, &request) ||
	 !outbuf_flush(&mxp_out) ||
	 !read_mxp_response(&mxp_in, &response)) {
      outbuf_printf(&err, "%s: unable to identify myself at %s\n",
	 cmdname, hptext);
      outbuf_flush(&err);
      exit(1);
   }
   if (response.status != MXP_SUCCESS) {
      outbuf_printf(&err, "%s: identification to %s as \"%.*s\" failed: %.*s\n",
	 cmdname, hptext, myname.len, myname.s,
	 response.message.len, response.message.s);
      outbuf_flush(&err);
      exit(1);
   }

   /* acquire lock */
   stralloc_copys(&request.keyword, "lock");
   stralloc_copys(&request.parameter, lockname);
   if (!write_mxp_request(&mxp_out, &request) || !outbuf_flush(&mxp_out)) {
      outbuf_printf(&err, "%s: unable to send lock request\n", cmdname);
      outbuf_flush(&err);
      exit(1);
   }
   do {
      if (!read_mxp_response(&mxp_in, &response)) {
	 outbuf_printf(&err, "%s: unable to receive response to lock command\n",
	    cmdname);
	 outbuf_flush(&err);
	 exit(1);
      }
   } while (response.status == MXP_CONTINUATION);
   if (response.status == MXP_FAILURE) {
      outbuf_printf(&err, "%s: lock of \"%s\" failed: %.*s\n",
	 cmdname, hptext, lockname,
	 response.message.len, response.message.s);
      outbuf_flush(&err);
      exit(1);
   }

   /* we have the lock acquired, execute command now */
   pid_t pid = fork();
   if (pid < 0) {
      outbuf_printf(&err, "%s: unable to fork: %s\n", cmdname, strerror(errno));
      outbuf_flush(&err);
      exit(1);
   }
   if (pid == 0) {
      close(fd);
      execvp(argv[0], argv);
      exit(255);
   }
   int wstat;
   if (waitpid(pid, &wstat, 0) < 0) {
      outbuf_printf(&err, "%s: unable to wait: %s\n", cmdname, strerror(errno));
      outbuf_flush(&err);
      exit(1);
   }

   /* lock is released by closing */
   close(fd);
}
